/*
 * Copyright (C) 2006  Emmanuel VARAGNAT <>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "common.h"
#include "scan.h"

#define SCAN_BUFF_SIZE  8192

static
long find_motif(const unsigned char *meule_de_foin, size_t size_meule,
		const unsigned char *aiguille,      size_t size_aiguille)
{
  long i;

  for(i = 0; i < (long)(size_meule - size_aiguille); i++) {
    if(memcmp(meule_de_foin + i, aiguille, size_aiguille) == 0)
      return i;
  }

  return -1;
}

static
int motif_search(const unsigned char *buffer,
		 unsigned int size,
		 unsigned int head_size,
		 off_t global_offset,
		 struct motif_infos *motif)
{
  unsigned int done;
  long pos;
  
  done = 0;
  
  while(done < (size + head_size)
	&&
	(pos = find_motif(buffer + done, ( size + head_size ) - done, motif->motif, motif->motif_len)) >= 0)
    {
      off_t offset = global_offset + ( done + pos ) - head_size;

      motif->handler(offset, motif->extra);
      
      done += pos + 1;
    }

  return 0;
}


void scan_for_motifs(int fd,
		     struct motif_infos *motifs)
{
  unsigned char buff[SCAN_BUFF_SIZE];
  unsigned int motif_max_len, copied;
  ssize_t n;
  off_t global_offset;
  struct motif_infos *p;

  if( lseek(fd, 0, SEEK_SET) == -1 )
    INTERNAL_ERROR_EXIT("lseek: ", strerror(errno));

  motif_max_len = 0;
  for(p = motifs; p->motif_len; p++)
    if(p->motif_len > motif_max_len)
      motif_max_len = p->motif_len;  

  assert(motif_max_len > 0);

  copied = 0;
  global_offset = 0;
  while( (n = read(fd, buff + copied, SCAN_BUFF_SIZE - copied)) > 0 ) {
    
    for(p = motifs; p->motif_len; p++)
      if( motif_search(buff, (unsigned int)n, copied, global_offset, p) == -1 )
	break;
    
    global_offset += n;
    
    memmove(buff, buff + copied + n - motif_max_len, motif_max_len);
    copied = motif_max_len;
  }
}
